"""
This script verifies whether UXsim outputs reasonable solutions for multiple route networks in various configurations.
Note that it uses random numbers for rouce choice behavior, so the results may vary slightly between runs.
"""

import pytest
import random
from collections import defaultdict
from uxsim import *
import pandas as pd

def equal_tolerance(val, check, rel_tol=0.1, abs_tol=0.0):
    if check == 0 and abs_tol == 0:
        abs_tol = 0.1
    return abs(val - check) <= abs(check*rel_tol) + abs_tol

@pytest.mark.flaky(reruns=5)
def test_2route_equal():
    tt1s = []
    tt2s = []
    vol1s = []
    vol2s = []
    ttas = []

    for i in range(10):
        W = World(
            name="",
            deltan=5, 
            tmax=2000, 
            print_mode=1, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", -1, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=1000, free_flow_speed=20, jam_density=0.2)
        link22 = W.addLink("link22", "mid2", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 1500, 0.8)

        W.exec_simulation()

        W.analyzer.print_simple_stats()

        #W.analyzer.time_space_diagram_traj_links([link11, link12])
        #W.analyzer.time_space_diagram_traj_links([link21, link22])
        W.analyzer.basic_analysis()

        df = W.analyzer.link_to_pandas()

        tt1 = df[df["link"].isin(("link11", "link12"))]["average_travel_time"].sum()
        tt2 = df[df["link"].isin(("link21", "link22"))]["average_travel_time"].sum()
        vol1 = df[df["link"].isin(("link11", "link12"))]["traffic_volume"].sum()/2
        vol2 = df[df["link"].isin(("link21", "link22"))]["traffic_volume"].sum()/2

        df2 = W.analyzer.od_to_pandas()
        tta = df2["average_travel_time"]

        tt1s.append(tt1)
        tt2s.append(tt2)   
        vol1s.append(vol1)
        vol2s.append(vol2)
        ttas.append(tta)
    
    print(f"{np.average(tt1s) = }, {np.average(tt2s) = }, {np.average(vol1s) = }, {np.average(vol2s) = }, {np.average(tt1s) = }, {np.average(ttas) = }")

    assert equal_tolerance(np.average(tt1s), np.average(tt2s))
    assert equal_tolerance(np.average(vol1s), np.average(vol2s))
    assert equal_tolerance(np.average(tt1s), 100)
    assert equal_tolerance(np.average(ttas), 100)
    assert equal_tolerance(np.average(vol1s)+np.average(vol2s), 1200)

@pytest.mark.flaky(reruns=5)
def test_2route_equal_deltan1():
    tt1s = []
    tt2s = []
    vol1s = []
    vol2s = []
    ttas = []

    for i in range(5):
        W = World(
            name="",
            deltan=1, 
            tmax=2000, 
            print_mode=1, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", -1, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=1000, free_flow_speed=20, jam_density=0.2)
        link22 = W.addLink("link22", "mid2", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 1500, 0.8)

        W.exec_simulation()

        W.analyzer.print_simple_stats()

        #W.analyzer.time_space_diagram_traj_links([link11, link12])
        #W.analyzer.time_space_diagram_traj_links([link21, link22])
        W.analyzer.basic_analysis()

        df = W.analyzer.link_to_pandas()

        tt1 = df[df["link"].isin(("link11", "link12"))]["average_travel_time"].sum()
        tt2 = df[df["link"].isin(("link21", "link22"))]["average_travel_time"].sum()
        vol1 = df[df["link"].isin(("link11", "link12"))]["traffic_volume"].sum()/2
        vol2 = df[df["link"].isin(("link21", "link22"))]["traffic_volume"].sum()/2

        df2 = W.analyzer.od_to_pandas()
        tta = df2["average_travel_time"]

        tt1s.append(tt1)
        tt2s.append(tt2)   
        vol1s.append(vol1)
        vol2s.append(vol2)
        ttas.append(tta)
    
    print(f"{np.average(tt1s) = }, {np.average(tt2s) = }, {np.average(vol1s) = }, {np.average(vol2s) = }, {np.average(tt1s) = }, {np.average(ttas) = }")

    assert equal_tolerance(np.average(tt1s), np.average(tt2s), rel_tol=0.2)
    assert equal_tolerance(np.average(vol1s), np.average(vol2s), rel_tol=0.2)
    assert equal_tolerance(np.average(tt1s), 100, rel_tol=0.2)
    assert equal_tolerance(np.average(ttas), 100, rel_tol=0.2)
    assert equal_tolerance(np.average(vol1s)+np.average(vol2s), 1200)

@pytest.mark.flaky(reruns=5)
def test_2route_equal_iterative_exec():
    tt1s = []
    tt2s = []
    vol1s = []
    vol2s = []
    ttas = []

    for i in range(10):
        W = World(
            name="",
            deltan=5, 
            tmax=2000, 
            print_mode=1, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", -1, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=1000, free_flow_speed=20, jam_density=0.2)
        link22 = W.addLink("link22", "mid2", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 1500, 0.8)

        while W.check_simulation_ongoing():
            W.exec_simulation(duration_t=random.randint(50, 200)) 

        W.analyzer.print_simple_stats()

        #W.analyzer.time_space_diagram_traj_links([link11, link12])
        #W.analyzer.time_space_diagram_traj_links([link21, link22])
        W.analyzer.basic_analysis()

        df = W.analyzer.link_to_pandas()

        tt1 = df[df["link"].isin(("link11", "link12"))]["average_travel_time"].sum()
        tt2 = df[df["link"].isin(("link21", "link22"))]["average_travel_time"].sum()
        vol1 = df[df["link"].isin(("link11", "link12"))]["traffic_volume"].sum()/2
        vol2 = df[df["link"].isin(("link21", "link22"))]["traffic_volume"].sum()/2

        df2 = W.analyzer.od_to_pandas()
        tta = df2["average_travel_time"]

        tt1s.append(tt1)
        tt2s.append(tt2)   
        vol1s.append(vol1)
        vol2s.append(vol2)
        ttas.append(tta)
    
    print(f"{np.average(tt1s) = }\n{np.average(tt2s) = }\n{np.average(vol1s) = }\n{np.average(vol2s) = }\n{np.average(tt1s) = }\n{np.average(ttas) = }")

    assert equal_tolerance(np.average(tt1s), np.average(tt2s))
    assert equal_tolerance(np.average(vol1s), np.average(vol2s))
    assert equal_tolerance(np.average(tt1s), 100)
    assert equal_tolerance(np.average(ttas), 100)
    assert equal_tolerance(np.average(vol1s)+np.average(vol2s), 1200)

@pytest.mark.flaky(reruns=5)
def test_2route_equal_but_different_structure():
    tt1s = []
    tt2s = []
    vol1s = []
    vol2s = []
    ttas = []

    for i in range(10):
        W = World(
            name="",
            deltan=5, 
            tmax=2000, 
            print_mode=1, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid21", -1, 1.5)
        W.addNode("mid22", -1, 0.5)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=2000, free_flow_speed=20, jam_density=0.2)
        link12 = W.addLink("link12", "mid1", "dest", length=2000, free_flow_speed=10, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid21", length=1000, free_flow_speed=10, jam_density=0.2)
        link22 = W.addLink("link22", "mid21", "mid22", length=1000, free_flow_speed=10, jam_density=0.2)
        link23 = W.addLink("link23", "mid22", "dest", length=2000, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 1500, 0.8)

        W.exec_simulation()

        W.analyzer.print_simple_stats()

        W.analyzer.basic_analysis()

        df = W.analyzer.link_to_pandas()
        tt1 = df[df["link"].isin(("link11", "link12"))]["average_travel_time"].sum()
        tt2 = df[df["link"].isin(("link21", "link22", "link23"))]["average_travel_time"].sum()
        vol1 = df[df["link"].isin(("link11", "link12"))]["traffic_volume"].sum()/2
        vol2 = df[df["link"].isin(("link21", "link22", "link23"))]["traffic_volume"].sum()/3

        df2 = W.analyzer.od_to_pandas()
        tta = df2["average_travel_time"]

        tt1s.append(tt1)
        tt2s.append(tt2)   
        vol1s.append(vol1)
        vol2s.append(vol2)
        ttas.append(tta)
        
    print(f"{np.average(tt1s) = }, {np.average(tt2s) = }, {np.average(vol1s) = }, {np.average(vol2s) = }, {np.average(tt1s) = }, {np.average(ttas) = }")

    assert equal_tolerance(np.average(tt1s), np.average(tt2s))
    assert equal_tolerance(np.average(vol1s), np.average(vol2s))
    assert equal_tolerance(np.average(tt1s), 300)
    assert equal_tolerance(np.average(ttas), 300)
    assert equal_tolerance(np.average(vol1s)+np.average(vol2s), 1200)

@pytest.mark.flaky(reruns=5)
def test_2route_one_is_too_long():
    tt1s = []
    tt2s = []
    vol1s = []
    vol2s = []
    ttas = []

    for i in range(3):
        W = World(
            name="",
            deltan=5, 
            tmax=2000, 
            print_mode=1, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", -1, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=2000, free_flow_speed=20, jam_density=0.2)
        link22 = W.addLink("link22", "mid2", "dest", length=2000, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 1500, 0.8)

        W.exec_simulation()

        W.analyzer.print_simple_stats()

        #W.analyzer.time_space_diagram_traj_links([link11, link12])
        #W.analyzer.time_space_diagram_traj_links([link21, link22])
        W.analyzer.basic_analysis()

        df = W.analyzer.link_to_pandas()

        tt1 = df[df["link"].isin(("link11", "link12"))]["average_travel_time"].sum()
        tt2 = df[df["link"].isin(("link21", "link22"))]["average_travel_time"].sum()
        vol1 = df[df["link"].isin(("link11", "link12"))]["traffic_volume"].sum()/2
        vol2 = df[df["link"].isin(("link21", "link22"))]["traffic_volume"].sum()/2

        df2 = W.analyzer.od_to_pandas()
        tta = df2["average_travel_time"]

        tt1s.append(tt1)
        tt2s.append(tt2)   
        vol1s.append(vol1)
        vol2s.append(vol2)
        ttas.append(tta)

    print(f"{np.average(tt1s) = }\n{np.average(tt2s) = }\n{np.average(vol1s) = }\n{np.average(vol2s) = }\n{np.average(tt1s) = }\n{np.average(ttas) = }")

    assert equal_tolerance(np.average(tt1s), 100)
    assert equal_tolerance(np.average(vol1s), 1200)
    assert equal_tolerance(np.average(vol2s), 0, abs_tol=100)
    assert equal_tolerance(np.average(ttas), 100)
    assert equal_tolerance(np.average(vol1s)+np.average(vol2s), 1200)

@pytest.mark.flaky(reruns=5)
def test_2route_one_is_too_long_but_u_changes_during_simulation():
    vol1_es = []
    vol2_es = []
    vol1_ls = []
    vol2_ls = []

    for i in range(10):
        W = World(
            name="",
            deltan=5, 
            tmax=2000, 
            print_mode=1, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", -1, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=2000, free_flow_speed=20, jam_density=0.2)
        link22 = W.addLink("link22", "mid2", "dest", length=2000, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 1500, 0.8)

        W.exec_simulation(duration_t=500)
        link11.change_free_flow_speed(5)
        link12.change_free_flow_speed(5)
        W.exec_simulation()
        

        W.analyzer.print_simple_stats()

        W.analyzer.basic_analysis()

        W.analyzer.compute_edie_state()

        vol1_es.append(link11.q_mat[:8,5].sum()*link11.edie_dt)
        vol1_ls.append(link11.q_mat[8:,5].sum()*link11.edie_dt)
        vol2_es.append(link21.q_mat[:8,5].sum()*link21.edie_dt)
        vol2_ls.append(link21.q_mat[8:,5].sum()*link21.edie_dt)

    print(f"{np.average(vol1_es) = }\n{np.average(vol1_ls) = }\n{np.average(vol2_es) = }\n{np.average(vol2_ls) = }")
    assert equal_tolerance(np.average(vol1_es), 500)
    assert equal_tolerance(np.average(vol1_ls), 20, abs_tol=100)
    assert equal_tolerance(np.average(vol2_es), 200)
    assert equal_tolerance(np.average(vol2_ls), 470)

@pytest.mark.flaky(reruns=5)
def test_2route_change_too_small_pricing():
    vol1_es = []
    vol2_es = []
    vol1_ls = []
    vol2_ls = []

    for i in range(10):
        W = World(
            name="",
            deltan=5, 
            tmax=4000, 
            print_mode=0, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100,
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", -1, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=2000, free_flow_speed=20, jam_density=0.2)
        link22 = W.addLink("link22", "mid2", "dest", length=2000, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 3000, 0.6)

        W.exec_simulation(duration_t=500)
        link11.route_choice_penalty = 90
        W.exec_simulation()
        
        W.analyzer.print_simple_stats()

        # W.analyzer.time_space_diagram_traj_links([link11, link12])
        # W.analyzer.time_space_diagram_traj_links([link21, link22])

        W.analyzer.basic_analysis()

        W.analyzer.compute_edie_state()

        vol1_es.append(link11.q_mat[:5,5].sum()*link11.edie_dt)
        vol1_ls.append(link11.q_mat[15:,5].sum()*link11.edie_dt)
        vol2_es.append(link21.q_mat[:5,5].sum()*link21.edie_dt)
        vol2_ls.append(link21.q_mat[15:,5].sum()*link21.edie_dt)

    print(f"{np.average(vol1_es) = }\n{np.average(vol1_ls) = }\n{np.average(vol2_es) = }\n{np.average(vol2_ls) = }")
    assert equal_tolerance(np.average(vol1_es), 340, abs_tol=100)
    assert equal_tolerance(np.average(vol1_ls), 750)
    assert equal_tolerance(np.average(vol2_es), 0, abs_tol=100)
    assert equal_tolerance(np.average(vol2_ls), 0)

@pytest.mark.flaky(reruns=5)
def test_2route_change_too_large_pricing_iterative():
    vol1_es = []
    vol2_es = []
    vol1_ls = []
    vol2_ls = []

    for i in range(10):
        W = World(
            name="",
            deltan=5, 
            tmax=4000, 
            print_mode=0, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100,
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", -1, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=2000, free_flow_speed=20, jam_density=0.2)
        link22 = W.addLink("link22", "mid2", "dest", length=2000, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 3000, 0.6)

        W.exec_simulation(duration_t=500)
        link11.route_choice_penalty = 110
        W.exec_simulation()
        
        W.analyzer.print_simple_stats()

        # W.analyzer.time_space_diagram_traj_links([link11, link12])
        # W.analyzer.time_space_diagram_traj_links([link21, link22])

        W.analyzer.basic_analysis()

        W.analyzer.compute_edie_state()

        vol1_es.append(link11.q_mat[:5,5].sum()*link11.edie_dt)
        vol1_ls.append(link11.q_mat[15:,5].sum()*link11.edie_dt)
        vol2_es.append(link21.q_mat[:5,5].sum()*link21.edie_dt)
        vol2_ls.append(link21.q_mat[15:,5].sum()*link21.edie_dt)

    print(f"{np.average(vol1_es) = }\n{np.average(vol1_ls) = }\n{np.average(vol2_es) = }\n{np.average(vol2_ls) = }")
    assert equal_tolerance(np.average(vol1_es), 340, abs_tol=100)
    assert equal_tolerance(np.average(vol1_ls), 0)
    assert equal_tolerance(np.average(vol2_es), 0, abs_tol=100)
    assert equal_tolerance(np.average(vol2_ls), 750)


@pytest.mark.flaky(reruns=5)
def test_route_choice_4route_congestion_avoidance():
    tt1s = []
    tt2s = []
    tt3s = []
    tt4s = []
    vol1s = []
    vol2s = []
    vol3s = []
    vol4s = []
    ttas = []

    for i in range(20):
        W = World(
            name="",
            deltan=5, 
            tmax=3000, 
            print_mode=1, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100,
            duo_update_weight=0.3
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", 2, 1)
        W.addNode("mid3", 3, 1)
        W.addNode("mid4", 4, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.3)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.3)
        link22 = W.addLink("link22", "mid2", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link31 = W.addLink("link31", "orig", "mid3", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.3)
        link32 = W.addLink("link32", "mid3", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link41 = W.addLink("link41", "orig", "mid4", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.3)
        link42 = W.addLink("link42", "mid4", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 2000, 0.8)

        W.exec_simulation() 

        W.analyzer.print_simple_stats()

        # W.analyzer.time_space_diagram_traj_links([link11, link12])
        # W.analyzer.time_space_diagram_traj_links([link21, link22])
        # W.analyzer.time_space_diagram_traj_links([link31, link32])
        # W.analyzer.time_space_diagram_traj_links([link41, link42])
        W.analyzer.basic_analysis()

        df = W.analyzer.link_to_pandas()

        tt1 = df[df["link"].isin(("link11", "link12"))]["average_travel_time"].sum()
        tt2 = df[df["link"].isin(("link21", "link22"))]["average_travel_time"].sum()
        tt3 = df[df["link"].isin(("link31", "link32"))]["average_travel_time"].sum()
        tt4 = df[df["link"].isin(("link41", "link42"))]["average_travel_time"].sum()
        vol1 = df[df["link"].isin(["link11"])]["traffic_volume"].values[0]
        vol2 = df[df["link"].isin(["link21"])]["traffic_volume"].values[0]
        vol3 = df[df["link"].isin(["link31"])]["traffic_volume"].values[0]
        vol4 = df[df["link"].isin(["link41"])]["traffic_volume"].values[0]

        df2 = W.analyzer.od_to_pandas()
        tta = df2["average_travel_time"].values[0]

        tt1s.append(tt1)
        tt2s.append(tt2)
        tt3s.append(tt3)
        tt4s.append(tt4)
        vol1s.append(vol1)
        vol2s.append(vol2)
        vol3s.append(vol3)
        vol4s.append(vol4)
            
        ttas.append(tta)

    print(f"{np.average(tt1s) = }\n{np.average(tt2s) = }\n{np.average(tt3s) = }\n{np.average(tt4s) = }\n{np.average(vol1s) = }\n{np.average(vol2s) = }\n{np.average(vol3s) = }\n{np.average(vol4s) = }\n{np.average(np.concatenate((tt1s, tt2s, tt3s, tt4s))) = }\n{np.average(np.concatenate((vol1s, vol2s, vol3s, vol4s))) = }")
    ttave = np.average(np.concatenate((tt1s, tt2s, tt3s, tt4s)))
    volave = np.average(np.concatenate((vol1s, vol2s, vol3s, vol4s)))

    assert equal_tolerance(np.average(tt1s), ttave, rel_tol=0.2)
    assert equal_tolerance(np.average(tt2s), ttave, rel_tol=0.2)
    assert equal_tolerance(np.average(tt3s), ttave, rel_tol=0.2)
    assert equal_tolerance(np.average(tt4s), ttave, rel_tol=0.2)
    assert equal_tolerance(np.average(vol1s), volave, rel_tol=0.2)
    assert equal_tolerance(np.average(vol2s), volave, rel_tol=0.2)
    assert equal_tolerance(np.average(vol3s), volave, rel_tol=0.2)
    assert equal_tolerance(np.average(vol4s), volave, rel_tol=0.2)
    assert equal_tolerance(volave*4, 2000*0.8)

def test_route_links_prefer():     
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid1",0,0)
    W.addNode("mid2",0,0)
    W.addNode("dest", 0, 0)
    link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20)
    link12 = W.addLink("link12", "orig", "mid1", length=1000, free_flow_speed=10)
    link21 = W.addLink("link21", "mid1", "mid2", length=1000, free_flow_speed=20)
    link22 = W.addLink("link22", "mid1", "mid2", length=1000, free_flow_speed=10)
    link31 = W.addLink("link31", "mid2", "dest", length=1000, free_flow_speed=20)
    link32 = W.addLink("link32", "mid2", "dest", length=1000, free_flow_speed=10)

    for t in range(0,1000, 10):
        W.addVehicle("orig", "dest", links_prefer=[link12, link21, link32], departure_time=t)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    df = W.analyzer.link_to_pandas()

    assert equal_tolerance(df[df["link"]=="link12"]["traffic_volume"].values[0], 500)
    assert equal_tolerance(df[df["link"]=="link21"]["traffic_volume"].values[0], 500)
    assert equal_tolerance(df[df["link"]=="link32"]["traffic_volume"].values[0], 500)

def test_route_links_avoid():
    W = World(
        name="",
        deltan=5,
        tmax=2000,
        print_mode=1, save_mode=1, show_mode=1,
        random_seed=None
    )

    W.addNode("orig", 0, 0)
    W.addNode("mid1",0,0)
    W.addNode("mid2",0,0)
    W.addNode("dest", 0, 0)
    link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20)
    link12 = W.addLink("link12", "orig", "mid1", length=1000, free_flow_speed=10)
    link21 = W.addLink("link21", "mid1", "mid2", length=1000, free_flow_speed=20)
    link22 = W.addLink("link22", "mid1", "mid2", length=1000, free_flow_speed=10)
    link31 = W.addLink("link31", "mid2", "dest", length=1000, free_flow_speed=20)
    link32 = W.addLink("link32", "mid2", "dest", length=1000, free_flow_speed=10)

    for t in range(0,1000, 10):
        W.addVehicle("orig", "dest", links_avoid=[link12, link21, link32], departure_time=t)

    W.exec_simulation()

    W.analyzer.print_simple_stats()

    df = W.analyzer.link_to_pandas()

    assert equal_tolerance(df[df["link"]=="link11"]["traffic_volume"].values[0], 500)
    assert equal_tolerance(df[df["link"]=="link22"]["traffic_volume"].values[0], 500)
    assert equal_tolerance(df[df["link"]=="link31"]["traffic_volume"].values[0], 500)

@pytest.mark.flaky(reruns=5)
def test_route_multiple_links_between_same_nodes():    
    vol11s = []
    vol12s = []
    vol21s = []
    vol22s = []

    for i in range(10):

        W = World(
            name="",
            deltan=5,
            tmax=2000,
            print_mode=0, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1",0,0)
        W.addNode("mid2",0,0)
        W.addNode("dest", 0, 0)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20)
        link12 = W.addLink("link12", "orig", "mid1", length=1000, free_flow_speed=10)
        link21 = W.addLink("link21", "orig", "mid2", length=1000, free_flow_speed=20)
        link22 = W.addLink("link22", "orig", "mid2", length=1000, free_flow_speed=10)
        link31 = W.addLink("link31", "mid1", "dest", length=1000, free_flow_speed=20)
        link32 = W.addLink("link32", "mid2", "dest", length=1000, free_flow_speed=20)

        W.adddemand("orig", "dest", 0, 1000, 0.8)

        W.exec_simulation()

        W.analyzer.print_simple_stats()

        df = W.analyzer.link_to_pandas()
        #display(df)

        vol11s.append(df[df["link"]=="link11"]["traffic_volume"].values[0])
        vol12s.append(df[df["link"]=="link12"]["traffic_volume"].values[0])
        vol21s.append(df[df["link"]=="link21"]["traffic_volume"].values[0])
        vol22s.append(df[df["link"]=="link22"]["traffic_volume"].values[0])

    print(f"{np.average(vol11s) = }\n{np.average(vol12s) = }\n{np.average(vol21s) = }\n{np.average(vol22s) = }")
    assert equal_tolerance(np.average(vol11s), np.average(vol12s), rel_tol=0.2)
    assert equal_tolerance(np.average(vol21s), np.average(vol22s), rel_tol=0.2)
    assert equal_tolerance(np.average(vol11s)+np.average(vol12s), np.average(vol21s)+np.average(vol22s), rel_tol=0.2)


@pytest.mark.flaky(reruns=5)
def test_route_choice_update_gradual():
    res = defaultdict(list)

    for i in range(10):
        W = World(
            name="",
            deltan=5, 
            tmax=2000, 
            print_mode=1, save_mode=0, show_mode=1,
            random_seed=None,
            duo_update_time=300,
            duo_update_weight=1,
            route_choice_update_gradual=True
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", -1, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=2000, free_flow_speed=30, jam_density=0.2, capacity_out=0.4)
        link12 = W.addLink("link12", "mid1", "dest", length=500, free_flow_speed=30, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=2000, free_flow_speed=20, jam_density=0.2)
        link22 = W.addLink("link22", "mid2", "dest", length=500, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 1500, 0.8)

        W.exec_simulation()

        W.analyzer.print_simple_stats()

        W.analyzer.basic_analysis()

        df = W.analyzer.link_cumulative_to_pandas()

        for link, t in [("link11", 250), ("link11", 500), ("link11", 750), ("link11", 1000), ("link21", 250), ("link21", 500), ("link21", 750), ("link21", 1000)]:
            res[link, t].append(df.loc[(df["link"] == link) & (df["t"] == t), "arrival_count"].item())

    for key in res:
        res[key] = np.mean(res[key])

    assert equal_tolerance(res["link11", 250], 200, rel_tol=0.1)
    assert equal_tolerance(res["link11", 500], 360, rel_tol=0.1)
    assert equal_tolerance(res["link11", 750], 430, rel_tol=0.1)
    assert equal_tolerance(res["link11",1000], 460, rel_tol=0.1)
    assert equal_tolerance(res["link21", 250], 0, abs_tol=20)
    assert equal_tolerance(res["link21", 500], 40, abs_tol=20, rel_tol=0.2)
    assert equal_tolerance(res["link21", 750], 170, rel_tol=0.1)
    assert equal_tolerance(res["link21",1000], 340, rel_tol=0.1)

    
@pytest.mark.flaky(reruns=5)
def test_route_choice_update_instant():
    res = defaultdict(list)

    for i in range(10):
        W = World(
            name="",
            deltan=5, 
            tmax=2000, 
            print_mode=1, save_mode=0, show_mode=1,
            random_seed=None,
            duo_update_time=300,
            duo_update_weight=1,
            route_choice_update_gradual=False
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", -1, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=2000, free_flow_speed=30, jam_density=0.2, capacity_out=0.4)
        link12 = W.addLink("link12", "mid1", "dest", length=500, free_flow_speed=30, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=2000, free_flow_speed=20, jam_density=0.2)
        link22 = W.addLink("link22", "mid2", "dest", length=500, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 1500, 0.8)

        W.exec_simulation()

        W.analyzer.print_simple_stats()

        W.analyzer.basic_analysis()

        df = W.analyzer.link_cumulative_to_pandas()

        for link, t in [("link11", 250), ("link11", 500), ("link11", 750), ("link11", 1000), ("link21", 250), ("link21", 500), ("link21", 750), ("link21", 1000)]:
            res[link, t].append(df.loc[(df["link"] == link) & (df["t"] == t), "arrival_count"].item())

    for key in res:
        res[key] = np.mean(res[key])

    assert equal_tolerance(res["link11", 250], 200, rel_tol=0.1)
    assert equal_tolerance(res["link11", 500], 240, rel_tol=0.1)
    assert equal_tolerance(res["link11", 750], 240, rel_tol=0.1)
    assert equal_tolerance(res["link11",1000], 320, rel_tol=0.1)
    assert equal_tolerance(res["link21", 250], 0, abs_tol=20)
    assert equal_tolerance(res["link21", 500], 160, abs_tol=20, rel_tol=0.2)
    assert equal_tolerance(res["link21", 750], 360, rel_tol=0.1)
    assert equal_tolerance(res["link21",1000], 480, rel_tol=0.1)


@pytest.mark.flaky(reruns=5)
def test_route_choice_4route_congestion_avoidance_gradual():
    
    tt1s = []
    tt2s = []
    tt3s = []
    tt4s = []
    vol1s = []
    vol2s = []
    vol3s = []
    vol4s = []
    ttas = []

    for i in range(20):
        W = World(
            name="",
            deltan=5, 
            tmax=3000, 
            print_mode=1, save_mode=1, show_mode=1,
            random_seed=None,
            duo_update_time=100,
            duo_update_weight=0.5,
            route_choice_update_gradual=True
        )

        W.addNode("orig", 0, 0)
        W.addNode("mid1", 1, 1)
        W.addNode("mid2", 2, 1)
        W.addNode("mid3", 3, 1)
        W.addNode("mid4", 4, 1)
        W.addNode("dest", 0, 2)
        link11 = W.addLink("link11", "orig", "mid1", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.3)
        link12 = W.addLink("link12", "mid1", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link21 = W.addLink("link21", "orig", "mid2", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.3)
        link22 = W.addLink("link22", "mid2", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link31 = W.addLink("link31", "orig", "mid3", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.3)
        link32 = W.addLink("link32", "mid3", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        link41 = W.addLink("link41", "orig", "mid4", length=1000, free_flow_speed=20, jam_density=0.2, capacity_out=0.3)
        link42 = W.addLink("link42", "mid4", "dest", length=1000, free_flow_speed=20, jam_density=0.2)
        W.adddemand("orig", "dest", 0, 2000, 0.8)

        W.exec_simulation() 

        W.analyzer.print_simple_stats()

        # W.analyzer.time_space_diagram_traj_links([link11, link12])
        # W.analyzer.time_space_diagram_traj_links([link21, link22])
        # W.analyzer.time_space_diagram_traj_links([link31, link32])
        # W.analyzer.time_space_diagram_traj_links([link41, link42])
        W.analyzer.basic_analysis()

        df = W.analyzer.link_to_pandas()

        tt1 = df[df["link"].isin(("link11", "link12"))]["average_travel_time"].sum()
        tt2 = df[df["link"].isin(("link21", "link22"))]["average_travel_time"].sum()
        tt3 = df[df["link"].isin(("link31", "link32"))]["average_travel_time"].sum()
        tt4 = df[df["link"].isin(("link41", "link42"))]["average_travel_time"].sum()
        vol1 = df[df["link"].isin(["link11"])]["traffic_volume"].values[0]
        vol2 = df[df["link"].isin(["link21"])]["traffic_volume"].values[0]
        vol3 = df[df["link"].isin(["link31"])]["traffic_volume"].values[0]
        vol4 = df[df["link"].isin(["link41"])]["traffic_volume"].values[0]

        df2 = W.analyzer.od_to_pandas()
        tta = df2["average_travel_time"].values[0]

        tt1s.append(tt1)
        tt2s.append(tt2)
        tt3s.append(tt3)
        tt4s.append(tt4)
        vol1s.append(vol1)
        vol2s.append(vol2)
        vol3s.append(vol3)
        vol4s.append(vol4)
            
        ttas.append(tta)

    print(f"{np.average(tt1s) = }\n{np.average(tt2s) = }\n{np.average(tt3s) = }\n{np.average(tt4s) = }\n{np.average(vol1s) = }\n{np.average(vol2s) = }\n{np.average(vol3s) = }\n{np.average(vol4s) = }\n{np.average(np.concatenate((tt1s, tt2s, tt3s, tt4s))) = }\n{np.average(np.concatenate((vol1s, vol2s, vol3s, vol4s))) = }")
    ttave = np.average(np.concatenate((tt1s, tt2s, tt3s, tt4s)))
    volave = np.average(np.concatenate((vol1s, vol2s, vol3s, vol4s)))

    assert equal_tolerance(np.average(tt1s), ttave, rel_tol=0.1)
    assert equal_tolerance(np.average(tt2s), ttave, rel_tol=0.1)
    assert equal_tolerance(np.average(tt3s), ttave, rel_tol=0.1)
    assert equal_tolerance(np.average(tt4s), ttave, rel_tol=0.1)
    assert equal_tolerance(np.average(vol1s), volave, rel_tol=0.1)
    assert equal_tolerance(np.average(vol2s), volave, rel_tol=0.1)
    assert equal_tolerance(np.average(vol3s), volave, rel_tol=0.1)
    assert equal_tolerance(np.average(vol4s), volave, rel_tol=0.1)
    assert equal_tolerance(volave*4, 2000*0.8)


"""
This script test essential Jupyter notebook demos, namely `demo_notebook_01en.ipynb`, `demo_notebook_01jp.ipynb`, and `demo_notebook_05en_for_google_colab.ipynb`, can run without error.
Implementation is awkward, but it works. Other jupyter notebook demos are optionally tested by `test_notebook_demos_extra.py`.
"""


import pytest
import nbformat
import os

from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors import CellExecutionError

from IPython.display import display


class SkipCellsPreprocessor(ExecutePreprocessor):
    """特定のキーワードを含むセルをスキップするプリプロセッサ"""
    
    def __init__(self, exception_words=None, **kwargs):
        super().__init__(**kwargs)
        self.exception_words = exception_words or []
    
    def preprocess_cell(self, cell, resources, cell_index):
        """セルを処理する前に、スキップすべきかどうかをチェック"""
        if cell.cell_type == "code":
            if any(word in cell.source for word in self.exception_words):
                print(f"Skipping cell {cell_index}: {cell.source[:50]}...")
                return cell, resources
        
        return super().preprocess_cell(cell, resources, cell_index)

def test_demo_notebook_01en():                        
    exception_words = ["ResultGUIViewer", "%matplotlib"]
    notebook_dir = "demos_and_examples"
    
    for notebook in os.listdir(notebook_dir):
        if notebook.endswith(".ipynb") and "demo_notebook_01en" in notebook:
            full_path = os.path.join(notebook_dir, notebook)
            
            with open(full_path, "r", encoding="utf-8") as f:
                nb = nbformat.read(f, as_version=4)
            
            # カスタムプリプロセッサを使用
            ep = SkipCellsPreprocessor(
                exception_words=exception_words,
                timeout=1800, 
                kernel_name='python3'
            )
            
            try:
                # ノートブックを実行
                print(f"Testing notebook: {notebook}")
                ep.preprocess(nb, {'metadata': {'path': notebook_dir}})
                print(f"Successfully executed all non-skipped cells in {notebook}")
            except CellExecutionError as e:
                # エラーが発生したセルの情報を出力
                print(f"Error executing notebook {notebook}")
                print(f"Error: {str(e)}")
                pytest.fail(f"Error executing notebook {notebook}: {str(e)}")

def test_demo_notebook_01jp():
    exception_words = ["ResultGUIViewer", "%matplotlib"]
    notebook_dir = "demos_and_examples"
    
    for notebook in os.listdir(notebook_dir):
        if notebook.endswith(".ipynb") and "demo_notebook_01jp" in notebook:
            full_path = os.path.join(notebook_dir, notebook)
            
            with open(full_path, "r", encoding="utf-8") as f:
                nb = nbformat.read(f, as_version=4)
            
            # カスタムプリプロセッサを使用
            ep = SkipCellsPreprocessor(
                exception_words=exception_words,
                timeout=1800, 
                kernel_name='python3'
            )
            
            try:
                # ノートブックを実行
                print(f"Testing notebook: {notebook}")
                ep.preprocess(nb, {'metadata': {'path': notebook_dir}})
                print(f"Successfully executed all non-skipped cells in {notebook}")
            except CellExecutionError as e:
                # エラーが発生したセルの情報を出力
                print(f"Error executing notebook {notebook}")
                print(f"Error: {str(e)}")
                pytest.fail(f"Error executing notebook {notebook}: {str(e)}")

def test_demo_notebook_05en():
    exception_words = ["ResultGUIViewer", "%matplotlib", "!pip"]
    notebook_dir = "demos_and_examples"
    
    for notebook in os.listdir(notebook_dir):
        if notebook.endswith(".ipynb") and "demo_notebook_05en_for_google_colab" in notebook:
            full_path = os.path.join(notebook_dir, notebook)
            
            with open(full_path, "r", encoding="utf-8") as f:
                nb = nbformat.read(f, as_version=4)
            
            # カスタムプリプロセッサを使用
            ep = SkipCellsPreprocessor(
                exception_words=exception_words,
                timeout=1800, 
                kernel_name='python3'
            )
            
            try:
                # ノートブックを実行
                print(f"Testing notebook: {notebook}")
                ep.preprocess(nb, {'metadata': {'path': notebook_dir}})
                print(f"Successfully executed all non-skipped cells in {notebook}")
            except CellExecutionError as e:
                # エラーが発生したセルの情報を出力
                print(f"Error executing notebook {notebook}")
                print(f"Error: {str(e)}")
                pytest.fail(f"Error executing notebook {notebook}: {str(e)}")


"""
This script tests whether the example scripts can run without errors.
"""

import pytest
import subprocess
import os

# Directory containing example scripts
examples_dir = 'demos_and_examples'
# Ignore list
nontest_files = [
    "example_14en_multiple_signals_Deep_Reinforcement_Learning_pytorch.py", #takes too much time. DQN can be tested by example_12en_* as well.
    "example_17en_result_GUI_viewer_sioux_falls.py", #GUI viewer is not testable in Github Actions
    "example_18en_result_GUI_viewer_grid.py", #GUI viewer is not testable in Github Actions
]

# Dynamically generate test cases for each example script
def pytest_generate_tests(metafunc):
    # List all .py files in the examples_dir
    example_scripts = [f for f in os.listdir(examples_dir) if f.endswith('.py') and not f in nontest_files]
    # If the test function expects an "example_script" argument, parametrize it
    if 'example_script' in metafunc.fixturenames:
        metafunc.parametrize('example_script', example_scripts)

def test_example_runs(example_script):
    """Test that a Python example script runs successfully."""
    # Build the script path
    script_path = os.path.join(examples_dir, example_script)
    
    if "streamlit" in example_script:
        # Run streamlit app
        cmd = ['streamlit', 'run', script_path]
         
        try:
            result = subprocess.run(cmd, capture_output=True, text=True, timeout=60)    #Streamlit app should always timeout if work properly
        except subprocess.TimeoutExpired:
            assert True

    else:
        # Regular Python script
        cmd = ['python', script_path]

        # Run the script as a separate process
        result = subprocess.run(cmd, capture_output=True, text=True)
        # Assert that the script ran successfully
        assert result.returncode == 0, f"Script {example_script} failed with output:\n{result.stdout}\n{result.stderr}"



"""
Submodule for handling taxis.
"""

import random, math
import pandas as pd

class TripRequest:
    """
    A class representing a trip request (or travelers, passengers, cargo, etc.).
    """
    def __init__(s, W, H, orig, dest, depart_time, name=None, attribute=None):
        """
        Initializes a trip request.

        Parameters
        ----------
        W : World
            The world object.
        H : TaxiHandler
            The TaxiHandler object.
        orig : str | Node
            The origin node of the trip request.
        dest : str | Node
            The destination node of the trip request.
        depart_time : float
            The time at which the trip request departs.
        name : any, optional
            The name of the trip request. Default is None.
        attribute : any
            An optional attribute of the trip request. This can be used to store any information by users' need.
        """
        
        s.W = W
        s.H = H
        s.orig = s.W.get_node(orig)
        s.dest = s.W.get_node(dest)
        s.depart_time = depart_time
        s.attribute = attribute
        s.name = name

        s.get_taxi_time = None
        s.arrival_time = None
        s.taxi = None

    def __repr__(s):
        return f"TravelRequest({s.name}: {s.orig}, {s.dest}, {s.depart_time})"

    def get_on_taxi(s):
        """
        The event of getting on a taxi. This is called via `Vehicle.event_node`.
        """
        #print(f"{s.W.TIME}: {s} is getting on a taxi {s.taxi}")        
        s.taxi.add_dest(s.dest)
        s.taxi.node_event[s.dest] = s.arrive_at_dest

        s.get_taxi_time = s.W.TIME
        
        del s.taxi.node_event[s.orig]
    
    def arrive_at_dest(s):
        """
        The event of arriving at the destination. This is called via `Vehicle.event_node`.
        """
        #print(f"{s.W.TIME}: {s} arrived at the destination {s.dest}")
        s.arrival_time = s.W.TIME
        
        del s.taxi.node_event[s.dest]

class TaxiHandler:
    """
    A base class for handling the assignment of trip requests to taxis (or ridesourcing, ridehailing, ridesharing, robot-taxi, shared mobility, whatever).
    """
    def __init__(s, W):
        """
        Initializes a taxi handler.

        Parameters
        ----------
        W : World
            The world object.
        """
        s.trip_requests = []
        s.trip_requests_all = []
        s.W = W
        random.seed(s.W.random_seed)

    def add_trip_request(s, orig, dest, depart_time, name=None, attribute=None):
        """
        Adds a trip request to this handler.

        Parameters
        ----------
        orig : str | Node
            The origin node of the trip request.
        dest : str | Node
            The destination node of the trip request.
        depart_time : float
            The time at which the trip request departs.
        name : any, optional
            The name of the trip request. Default is None.
        attribute : any
            An optional attribute of the trip request. This can be used to store any information by users' need.
        """
        if name == None:
            name = len(s.trip_requests)
        s.trip_requests.append(TripRequest(s.W, s, orig, dest, depart_time, name, attribute=attribute))
        s.trip_requests_all.append(s.trip_requests[-1])
    
    def assign_taxi(s, vehicle, trip_request):
        """
        Assigns a trip request to a vehicle.

        Parameters
        ----------
        vehicle : Vehicle
            The vehicle to assign the trip request to.
        """
        #print(f"{s.W.TIME}: Assigning {trip_request} to {vehicle}")

        if trip_request not in s.trip_requests:
            raise ValueError("Travel request not found in the list of trip requests")

        vehicle.add_dest(trip_request.orig)
        vehicle.node_event[trip_request.orig] = trip_request.get_on_taxi
        trip_request.taxi = vehicle
        s.trip_requests.remove(trip_request)
    
    def get_trip(s, tr):
        """
        Get a TripRequest instance by name or object.

        Parameters
        ----------
        tr : any | TripRequest
            The name or object of the trip request.
        """
        if type(tr) is TripRequest:
            if tr in s.trip_requests_all:
                return tr
        else:
            for trr in s.trip_requests_all:
                if tr == trr.name:
                    return trr
        raise Exception(f"'{tr}' is not TripRequest of this TaxiHandler")


    def compute_stats(s):
        """
        Computes basic statistics of the taxi travels.
        """

        s.n_total_requests = len(s.trip_requests_all)*s.W.DELTAN
        s.n_completed_requests = 0
        s.waiting_times = []
        s.travel_times = []
        s.invehicle_times = []
        s.number_of_taxis = sum([1 for veh in s.W.VEHICLES.values() if veh.mode == "taxi"])*s.W.DELTAN
        for tr in s.trip_requests_all:
            if tr.arrival_time != None:
                s.n_completed_requests += s.W.DELTAN
                s.invehicle_times.append(tr.arrival_time - tr.get_taxi_time)
                s.travel_times.append(tr.arrival_time - tr.depart_time)
            if tr.get_taxi_time != None:
                s.waiting_times.append(tr.get_taxi_time - tr.depart_time)
    
    def basic_to_pandas(s):
        """
        Converts the basic statistics on taxi travels to a pandas DataFrame.
        """
        s.compute_stats()
        data = {
            "total_trips": [s.n_total_requests],
            "completed_trips": [s.n_completed_requests],
            "average_waiting_time": [sum(s.waiting_times)/len(s.waiting_times) if len(s.waiting_times) > 0 else 0],
            "average_invehicle_time": [sum(s.invehicle_times)/len(s.invehicle_times) if len(s.invehicle_times) > 0 else 0],
            "average_travel_time": [sum(s.travel_times)/len(s.travel_times) if len(s.travel_times) > 0 else 0],
        }
        return pd.DataFrame(data)
    
    def print_stats(s):
        """
        Prints the basic statistics.
        """
        s.compute_stats()
        print("results for taxi transportation:")
        print(f" total trip requests: {s.n_total_requests}")
        print(f" completed trip requests: {s.n_completed_requests}")
        print(f" completed trip requests ratio: {s.n_completed_requests/s.n_total_requests if s.n_total_requests > 0 else 0: .2f}")
        print(f" average number of completed requests per taxi: {s.n_completed_requests/s.number_of_taxis: .2f}")
        print(f" average waiting time: {sum(s.waiting_times)/len(s.waiting_times) if len(s.waiting_times) > 0 else 0: .1f}")
        print(f" average in-vehicle time: {sum(s.invehicle_times)/len(s.invehicle_times) if len(s.invehicle_times) > 0 else 0: .1f}")
        print(f" average trip time: {sum(s.travel_times)/len(s.travel_times) if len(s.travel_times) > 0 else 0: .1f}")
    
    def trips_to_pandas(s):
        """
        Converts the trips and their travel records to a pandas DataFrame.
        """
        s.compute_stats()
        data = {
            "orig": [tr.orig.name for tr in s.trip_requests_all],
            "dest": [tr.dest.name for tr in s.trip_requests_all],
            "depart_time": [tr.depart_time for tr in s.trip_requests_all],
            "get_taxi_time": [tr.get_taxi_time for tr in s.trip_requests_all],
            "arrival_time": [tr.arrival_time for tr in s.trip_requests_all],
            "travel_time": [tr.arrival_time - tr.depart_time if tr.arrival_time != None else None for tr in s.trip_requests_all],
            "waiting_time": [tr.get_taxi_time - tr.depart_time if tr.get_taxi_time != None else None for tr in s.trip_requests_all],
            "used_taxi": [tr.taxi.name if tr.taxi != None else None for tr in s.trip_requests_all],
        }
        return pd.DataFrame(data)


class TaxiHandler_random(TaxiHandler):
    """
    A naive taxi handler that assigns trip requests to random taxis. 
    """
    def __init__(s, W):
        super().__init__(W)
    
    def assign_trip_request_to_taxi(s):
        """
        Assigns trip request to random available taxi.
        """
        vacant_taxis = [veh for veh in s.W.VEHICLES.values() if veh.mode == "taxi" and veh.state == "run" and veh.dest == None]
        random.shuffle(vacant_taxis)
        for trip_request in s.trip_requests[:]:
            if len(vacant_taxis) == 0:
                break
            if trip_request.depart_time <= s.W.TIME:
                taxi = vacant_taxis.pop()
                s.assign_taxi(taxi, trip_request)


class TaxiHandler_nearest(TaxiHandler):
    """
    A taxi handler that assigns trip requests to nearest taxis (based on Euclidean distance). 
    """
    def __init__(s, W):
        super().__init__(W)
    
    def assign_trip_request_to_taxi(s):
        """
        Assigns trip request to nearest available taxi.
        """
        vacant_taxis = [veh for veh in s.W.VEHICLES.values() if veh.mode == "taxi" and veh.state == "run" and veh.dest == None]
        for trip_request in s.trip_requests[:]:
            if len(vacant_taxis) == 0:
                break
            if trip_request.depart_time <= s.W.TIME:
                dist_tmp = float("inf")
                taxi_tmp = None
                for taxi in vacant_taxis[:]:
                    x0, y0 = taxi.get_xy_coords()
                    x1, y1 = trip_request.orig.x, trip_request.orig.y
                    dist = math.sqrt((x0-x1)**2 + (y0-y1)**2)
                    if dist <= dist_tmp:
                        dist_tmp = dist
                        taxi_tmp = taxi
                if taxi_tmp != None:
                    vacant_taxis.remove(taxi_tmp)
                    s.assign_taxi(taxi_tmp, trip_request)


class TaxiHandler_nearest_matching_radious(TaxiHandler):
    """
    A taxi handler that assigns trip requests to nearest taxis that are within a certain radious of the origin node (based on Euclidean distance. 
    """
    def __init__(s, W, matching_radious):
        super().__init__(W)
        s.matching_radious = matching_radious
    
    def assign_trip_request_to_taxi(s):
        """
        Assigns trip request to nearest available taxi that is within the radious of the origin node.
        """
        vacant_taxis = [veh for veh in s.W.VEHICLES.values() if veh.mode == "taxi" and veh.state == "run" and veh.dest == None]
        for trip_request in s.trip_requests[:]:
            if len(vacant_taxis) == 0:
                break
            if trip_request.depart_time <= s.W.TIME:
                dist_tmp = float("inf")
                taxi_tmp = None
                for taxi in vacant_taxis[:]:
                    x0, y0 = taxi.get_xy_coords()
                    x1, y1 = trip_request.orig.x, trip_request.orig.y
                    dist = math.sqrt((x0-x1)**2 + (y0-y1)**2)
                    if dist <= s.matching_radious and dist <= dist_tmp:
                        dist_tmp = dist
                        taxi_tmp = taxi
                if taxi_tmp != None:
                    vacant_taxis.remove(taxi_tmp)
                    s.assign_taxi(taxi_tmp, trip_request)


class TaxiHandler_nearest_network_distance(TaxiHandler):
    """
    A taxi handler that assigns trip requests to nearest taxis based on network distance considering travel time as well. 
    This would be useful, but not implemented.
    """
    pass